from pyblast.constants import Constants as C
from uuid import uuid4
from pyblast.exceptions import SeqRecordValidationError
import networkx as nx
from pyblast.utils import is_circular


class GraphDB(object):
    def __init__(self):
        self.graph = nx.DiGraph()
        self.mapping = {}

    def new_key(self):
        """Generate a new unique key"""
        return str(uuid4())

    def key(self, object):
        """Model id to key. To ensure uniqueness of the models being added."""
        k = self.mapping.get(id(object), None)
        return k

    def get_many(self, keys, default=C.NULL):
        """Get many objects from list of keys."""
        records = []
        for k in keys:
            records.append(self.get(k, default=default))
        return records

    def get(self, key, default=C.NULL):
        """Get object from a key"""
        if default is not C.NULL:
            if key not in self.graph:
                return default
            return self.graph.node[key]["object"]
        return self.graph.node[key]["object"]

    def add(self, object):
        """Add an object"""
        if self.key(object):
            return self.key(object)
        else:
            key = self.new_key()
            self.graph.add_node(key, object=object)
            self.mapping[id(object)] = key
            return key

    def add_many(self, objects):
        """Add many objects"""
        keys = []
        for o in objects:
            keys.append(self.add(o))
        return keys

    def add_edge(self, k1, k2, **kwargs):
        """Add an edge"""
        return self.graph.add_edge(k1, k2, **kwargs)

    def to_dict(self):
        """Return key to object dictionary"""
        d = {}
        for key in self.graph:
            d[key] = self.get(key)
        return d


class SeqRecordDB(GraphDB):
    @staticmethod
    def is_circular(records):
        return is_circular(records)

    @property
    def records(self):
        return self.to_dict()

    @classmethod
    def validate_records(cls, records):
        for r in records:
            cls.validate_circular(r)

    @staticmethod
    def validate_circular(r):
        valid_keys = [C.CIRCULAR, C.LINEAR, C.TOPOLOGY]
        if all([k not in r.annotations for k in valid_keys]):
            raise SeqRecordValidationError(
                "SeqRecord {} is missing a any of the valid keys in its annotation: {keys}. This must be provided.".format(
                    r, keys=valid_keys
                )
            )
        if C.LINEAR in r.annotations and C.CIRCULAR in r.annotations:
            if r.annotations[C.LINEAR] is r.annotations[C.CIRCULAR]:
                raise SeqRecordValidationError(
                    "SeqRecord {} has conflicting topology definitions for '{linear}' and '{circular}'".format(
                        r, circular=C.CIRCULAR, linear=C.LINEAR
                    )
                )

    def transform(self, parent_key, transform, transform_label):
        record = self.get(parent_key)
        new_record = transform(record)
        new_key = self.add(new_record)
        self.add_edge(
            parent_key,
            new_key,
            **{C.TRANSFORMATION: transform_label, C.PARENT: parent_key}
        )
        new_record.id = new_key
        return new_key

    def add_with_transformation(self, record, transform, transform_label):
        key = self.add(record)
        return self.transform(key, transform, transform_label)

    def add_many_with_transformations(self, records, transform, transform_label):
        return [
            self.add_with_transformation(r, transform, transform_label) for r in records
        ]

    def get_origin_key(self, key):
        tree = nx.bfs_tree(self.graph, key, reverse=True)
        roots = [n for n, d in dict(tree.out_degree()).items() if d == 0]
        return roots[0]

    def get_origin(self, key):
        return self.get(self.get_origin_key(key))

    def add(self, record, validate=True):
        if self.key(record):
            return self.key(record)
        else:
            if validate:
                self.validate_records([record])
            return super().add(record)

    def add_many(self, records):
        self.validate_records(records)
        return super().add_many(records)



class Constants(object):

    PARENT = "origin_record"
    CIRCULAR = "circular"
    LINEAR = "linear"
    TRANSFORMATION = "modification"
    OLD_KEY = "original_record_id"
    PSEUDOCIRCULAR = "pseudocircularized"
    COPY_RECORD = "copy_record"
    TOPOLOGY = "topology"

    class NULL(object):
        """A not NONE class."""



from Bio.SeqRecord import SeqRecord
from Bio.SeqFeature import SeqFeature, FeatureLocation
from uuid import uuid4
import re
from pyblast.utils import new_feature_location


class BlastResultParser(object):
    @staticmethod
    def str_to_f_to_i(v):
        """"""
        try:
            v = float(v)
        except ValueError:
            pass
        try:
            v = int(v)
        except ValueError:
            pass
        return v

    @staticmethod
    def _extract_metadata(r, delim):
        """Extracts information from the raw text file BLAST produces"""
        g = re.search(
            "#\s*(?P<blast_ver>.+)\n"
            + "# Query:\s*(?P<query>.*)\n"
            + "# Database:\s*(?P<database>.+)\n"
            + "(?:# Fields:\s*(?P<fields>.+))?",
            r,
        )
        metadata = g.groupdict()
        if metadata["fields"] is None:
            return metadata
        fields_array = re.split("\s*{}\s*".format(delim), metadata["fields"])
        metadata["fields"] = fields_array
        return metadata

    @staticmethod
    def _get_alignment_rows(r):
        """Split text into alignment rows"""
        return re.findall("\n([^#].*)", r)

    @classmethod
    def _validate_matches(cls, raw_matches, fields):
        """Create a dictionary from the fields and rows"""
        match_dicts = []
        for m in raw_matches:
            values = [cls.str_to_f_to_i(v) for v in m.split("\t")]
            match_dicts.append(dict(list(zip(fields, values))))
        return match_dicts

    @staticmethod
    def __convert_strand_label(strand_lable):
        if strand_lable.strip().lower() != "plus":
            return -1
        return 1

    @classmethod
    def __clean_json(cls, data_list):
        for data in data_list:
            query = {}
            subject = {}

            query["start"] = data["q. start"]
            query["end"] = data["q. end"]
            query["bases"] = data["query seq"]
            query["strand"] = cls.__convert_strand_label(
                data.get("query strand", "plus")
            )
            query["length"] = data["query length"]
            query["sequence_id"] = data["query acc."]

            subject["start"] = data["s. start"]
            subject["end"] = data["s. end"]
            subject["bases"] = data["subject seq"]
            subject["strand"] = cls.__convert_strand_label(data["subject strand"])
            subject["length"] = data["subject length"]
            subject["sequence_id"] = data["subject acc."]

            meta = dict(data)
            yield {"query": query, "subject": subject, "meta": meta}

    @classmethod
    def raw_results_to_json(cls, raw_text, delim=","):
        """
        Converts raw BLAST text into a flatten dictionary

        :param raw_text: raw text from BLAST results
        :type raw_text: basestring
        :param delim: delimiter for parsing
        :type delim: basestring
        :return: flattened dictionary
        :rtype: dict
        """

        # print(results)
        if raw_text.strip() == "":
            return []
        meta = cls._extract_metadata(raw_text, delim)
        fields = meta["fields"]
        if fields is None:
            return []
        alignment_rows = cls._get_alignment_rows(raw_text)
        match_dicts = cls._validate_matches(alignment_rows, tuple(fields))
        data = list(cls.__clean_json(match_dicts))
        return data

    @staticmethod
    def alignment_to_feature(align: dict, type: str, id: str):
        return SeqFeature(
            location=new_feature_location(
                align["start"] - 1,
                align["end"],
                strand=align["strand"],
                length=align["length"],
            ),
            type=type,
            id=id,
        )

    @classmethod
    def alignment_to_seq_record(cls, align, name, id, seq_dict: dict):

        origin_record = seq_dict[align["sequence_id"]]

        new_record = SeqRecord(
            id=id,
            seq=origin_record.seq,
            name=name,
            features=[cls.alignment_to_feature(align, "alignment", id)],
            annotations={},
        )

    @staticmethod
    def alignment_to_seqrecord(align: dict, seq_dict: dict):

        query_record = seq_dict[align["query"]["sequence_id"]]
        subject_record = seq_dict[align["subject"]["sequence_id"]]

        query_alignment = SeqRecord(
            seq=query_record.seq,
            id=str(uuid4()),
            name="QUERY__{}__alignedto__{}".format(
                subject_record.name, query_record.name
            ),
            features=[
                SeqFeature(
                    location=new_feature_location(
                        align["query"]["start"] - 1,
                        align["query"]["end"],
                        strand=1,
                        length=align["query"]["length"],
                    ),
                    type="alignment",
                    id="query alignment",
                )
            ],
            annotations={
                "alignment": align,
                "circular": query_record.annotations["circular"],
            },
        )
        if align["subject"]["strand"] != "plus":
            strand = -1
        else:
            strand = 1
        subject_alignment = SeqRecord(
            seq=subject_record.seq,
            id=str(uuid4()),
            name="SUBJECT__{}__alignedto__{}".format(
                subject_record.name, query_record.name
            ),
            features=[
                SeqFeature(
                    location=new_feature_location(
                        align["subject"]["start"] - 1,
                        align["subject"]["end"],
                        strand=strand,
                        length=align["subject"]["length"],
                    ),
                    type="alignment",
                    id="subject alignment",
                )
            ],
            annotations={
                "alignment": align,
                "circular": subject_record.annotations["circular"],
            },
        )

        return query_alignment, subject_alignment

    @staticmethod
    def get_perfect(data):
        """
        Returns only exact matches.

        :return:
        :rtype:
        """
        no_gaps = lambda x: x["meta"]["gaps"] == 0
        no_gap_opens = lambda x: x["meta"]["gap opens"] == 0
        identical = lambda x: x["meta"]["identical"] == x["meta"]["alignment length"]
        perfect = lambda x: all([no_gaps(x), no_gap_opens(x), identical(x)])
        return [r for r in data if perfect(r)]

    @staticmethod
    def get_with_perfect_subjects(data):
        """
        Returns only parsed alignments with 100% of the subject aligning to the query

        :return: perfect alignments
        :rtype:
        """
        f = lambda x: x["meta"]["alignment_length"] == x["subject"]["length"]
        return [r for r in data if f(r)]



import pprint
import ftplib
import subprocess
import os
import re
import shutil
from glob import glob
from termcolor import cprint, colored


class BlastWrapper(object):
    """A really shallow wrapper for running blast commands"""

    def __init__(self, path=None):
        if path is None:
            path = self.find_global_installations()
        if path is None:
            paths = self.find_bin_installations()
            if len(paths) > 0:
                path = paths[0]
                print('Found previous installation at "{}"'.format(path))
        self.path = path
        if path is not None:
            self._add_path_to_environment(path)

    @staticmethod
    def find_bin_installations():
        here = os.path.dirname(os.path.abspath(__file__))
        glob_pattern = os.path.join(here, "bin", "ncbi*/bin")
        dirs = sorted(list(glob(glob_pattern)))
        dirs = [d for d in dirs if os.path.isdir(os.path.abspath(d))]
        return dirs

    @staticmethod
    def find_global_installations():
        install_path = shutil.which("makeblastdb")
        if install_path is None:
            return None
        return os.path.split(install_path)[0]

    @staticmethod
    def _add_path_to_environment(path):
        """Add the blast installations located in 'bin' to the environment temporarily"""
        real_path = os.path.abspath(path)
        os.environ["PATH"] += ":" + real_path

    def is_installed(self):
        return shutil.which("makeblastdb")

    def install(self, email, platform):
        """
        Installs blast to bin directory.

        :param email: user's email
        :type email: basestring
        :param platform: platform (mac, linux, etc.)
        :type platform: basestring
        :return: None
        :rtype: None
        """
        dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "bin")
        if not os.path.isdir(dir):
            os.makedirs(dir)
        path = self.install_blast_using_ftp(email, platform, dir)
        self.path = os.path.join(path, "bin")

    @staticmethod
    def _platform_formats():
        version_pattern = "\d+\.\d+\.\d+\+"
        blast_pattern = "(ncbi-blast-{ver})-{platform}{ext}"
        tarball_ext = "\.tar\.gz"
        tarball_formats = {
            "mac": blast_pattern.format(
                ver=version_pattern, platform=".*?macosx.*?", ext=tarball_ext
            ),
            "linux(pentium)": blast_pattern.format(
                ver=version_pattern, platform="ia32-linux", ext=tarball_ext
            ),
            "linux(X64 chip)": blast_pattern.format(
                ver=version_pattern, platform="x64-linux", ext=tarball_ext
            ),
        }
        return tarball_formats

    @classmethod
    def _valid_platforms(cls):
        return list(cls._platform_formats().keys())

    @classmethod
    def _get_format(cls, platform):
        tarball_formats = cls._platform_formats()
        fmt = tarball_formats.get(platform, None)
        if fmt is None:
            raise Exception(
                'Cannot find platform "{}". Please select from {}'.format(
                    platform, list(tarball_formats.keys())
                )
            )
        return fmt

    @classmethod
    def install_blast_using_ftp(
        cls,
        email,
        platform,
        dir,
        user="anonymous",
        domain="ftp.ncbi.nlm.nih.gov",
        cwd="blast/executables/blast+/LATEST",
    ):
        # print config
        if not os.path.isdir(os.path.abspath(dir)):
            raise NotADirectoryError('Directory "{}" is not a directory'.format(dir))

        fmt = cls._get_format(platform)
        print("ftp config:")
        pp = pprint.PrettyPrinter(indent=4)
        pp.pprint(
            {
                "email": email,
                "platform": platform,
                "fileformat": fmt,
                "dir": dir,
                "user": user,
                "domain": domain,
                "cwd": cwd,
            }
        )
        print()

        # login
        ftp = ftplib.FTP(domain)
        ftp.login(user, email)

        # get files
        data = []
        ftp.cwd(cwd)
        ftp.dir(data.append)

        # find filename with format
        filename = None
        blastver = None
        for line in data:
            m = re.search(fmt, line)
            if m:
                filename = m.group()
                blastver = m.group(1)
                break
        if filename is None:
            raise Exception(
                "Unable to find blast binary. Data retrieved:\n{}".format(data)
            )

        input_file = os.path.join(dir, filename)

        # download tarball
        if not os.path.isfile(input_file):
            if not os.path.isdir(blastver):
                print("Downloading {}".format(filename))
                with open(input_file, "wb") as handle:
                    ftp.retrbinary("RETR {}".format(filename), handle.write)
                    ftp.quit()
                print("Download complete")

        # unzip and remove tarball
        if os.path.isfile(input_file):
            subprocess.call(["tar", "zxvpf", input_file, "-C", dir])
            os.remove(input_file)
        print("Blast installed to {}".format(input_file))
        return input_file

    def ask_to_install(self):
        msg = colored(
            "Blast not installed and so script cannot run. Would you like to install it now?\n",
            "red",
        )
        msg += colored(
            "Note that this will require an internet connection and your email to download blast "
            + "from NCBI (y|n): ",
            "green",
        )

        response = input(msg)
        if response == "y":
            email = input("email address: ")
            platform = input("platform ({}): ".format(BlastWrapper._valid_platforms()))
            self.install(email, platform)
        else:
            cprint(
                "User canceled. Blast not installed.\nPlease run 'pyblast install [EMAIL] [PLATFORM]' from the commandline."
                + "\nFor help, type 'pyblast install -- --help' in the commandline.",
                "red",
            )



"""AlignmentResults"""

import re
import json
from pyblast.schema import (
    QuerySchema,
    SubjectSchema,
    AlignmentSchema,
    AlignmentMetaSchema,
)


def str_to_f_to_i(v):
    """"""
    try:
        v = float(v)
    except ValueError:
        pass
    try:
        v = int(v)
    except ValueError:
        pass
    return v


class AlignmentResults(object):
    """Alignments results container"""

    def __init__(self, alignments):
        self._alignments = tuple(alignments)

    def __len__(self):
        return len(self._alignments)

    @property
    def alignments(self):
        return self._alignments

    @staticmethod
    def _extract_metadata(r, delim):
        """Extracts information from the raw text file BLAST produces"""
        g = re.search(
            "#\s*(?P<blast_ver>.+)\n"
            + "# Query:\s*(?P<query>.*)\n"
            + "# Database:\s*(?P<database>.+)\n"
            + "(?:# Fields:\s*(?P<fields>.+))?",
            r,
        )
        metadata = g.groupdict()
        if metadata["fields"] is None:
            return metadata
        fields_array = re.split("\s*{}\s*".format(delim), metadata["fields"])
        metadata["fields"] = fields_array
        return metadata

    @staticmethod
    def _get_alignment_rows(r):
        """Split text into alignment rows"""
        return re.findall("\n([^#].*)", r)

    @classmethod
    def _validate_matches(cls, raw_matches, fields):
        """Create a dictionary from the fields and rows"""
        match_dicts = []
        for m in raw_matches:
            values = [str_to_f_to_i(v) for v in m.split("\t")]
            match_dicts.append(dict(list(zip(fields, values))))
        return match_dicts

    @classmethod
    def _cleanup_raw_results(cls, raw_text, delim):
        """
        Converts raw BLAST text into a flatten dictionary

        :param raw_text: raw text from BLAST results
        :type raw_text: basestring
        :param delim: delimiter for parsing
        :type delim: basestring
        :return: flattened dictionary
        :rtype: dict
        """

        # print(results)
        if raw_text.strip() == "":
            return {}
        meta = cls._extract_metadata(raw_text, delim)
        fields = meta["fields"]
        if fields is None:
            return [{}]
        alignment_rows = cls._get_alignment_rows(raw_text)
        match_dicts = cls._validate_matches(alignment_rows, tuple(fields))
        return match_dicts

    @staticmethod
    def _serialize_data(data, context):
        """
        Serializes a flat dictionary of data into Query, Subject, AlginmentMeta, and Alignment models

        :param data: flatten dictionary
        :type data: dict
        :return: serialized data
        :rtype: dict
        """

        query_schema = QuerySchema(many=True)
        subject_schema = SubjectSchema(many=True)
        meta_schema = AlignmentMetaSchema(many=True)
        alignment_schema = AlignmentSchema(many=True)
        alignment_schema.context = context

        # load marshall alignment
        queries = query_schema.load(data)
        subjects = subject_schema.load(data)
        metas = meta_schema.load(data)

        serialized = alignment_schema.dump(
            [
                {"query": q, "subject": s, "meta": m}
                for q, s, m in zip(queries, subjects, metas)
            ]
        )
        return serialized

    @classmethod
    def parse_results(cls, data, delim, context=None):
        """Convert raw BLAST results into Alignment model"""
        # first clean up the data and get the flattened dictionary
        cleaned = cls._cleanup_raw_results(data, delim)
        if cleaned == [{}]:
            return cls(alignments=list())
        # force data to deserialized according to the schemas
        # if context is None:
        #     context = {'db': {}}
        serialized = cls._serialize_data(cleaned, context)

        # force data back into models
        alignment_schema = AlignmentSchema(many=True)
        alignment_schema.context = context
        deserialized = alignment_schema.load(serialized)

        return cls(alignments=deserialized)

    def get_perfect(self):
        """
        Returns only exact matches.

        :return:
        :rtype:
        """
        no_gaps = lambda x: x["meta"]["gaps"] == 0
        no_gap_opens = lambda x: x["meta"]["gaps_open"] == 0
        identical = lambda x: x["meta"]["identical"] == x["meta"]["alignment_length"]
        perfect = lambda x: all([no_gaps(x), no_gap_opens(x), identical(x)])
        return self.__class__([r for r in self.alignments if perfect(r)])

    def get_with_perfect_subjects(self):
        """
        Returns only parsed alignments with 100% of the subject aligning to the query

        :return: perfect alignments
        :rtype:
        """
        f = lambda x: x["meta"]["alignment_length"] == x["subject"]["length"]
        return self.__class__([r for r in self.alignments if f(r)])

    def to_json(self, path=None):
        if not path:
            return list(self.alignments)
        with open(path, "w") as out:
            json.dump(self.alignments, out)



"""
blast.py

More information: https://www.ncbi.nlm.nih.gov/guide/howto/run-blast-local/
BLAST User Manual: https://www.ncbi.nlm.nih.gov/books/NBK1762/
"""

import json
import os
import tempfile
from uuid import uuid4
from pyblast.utils import run_cmd
from pyblast.exceptions import PyBlastException
from pyblast.blast_bin import BlastWrapper
from Bio.SeqRecord import SeqRecord
import typing
from copy import deepcopy
from Bio import SeqIO
from pyblast.constants import Constants as C
from .blast_parser import BlastResultParser
from .json_parser import JSONParser
from pyblast.utils import glob_fasta_to_tmpfile, records_to_tmpfile, clean_records
from .seqdb import SeqRecordDB
from more_itertools import unique_everseen


class BlastBase(object):
    """
    A Blast initializer for running blast searches against subjects contained in a directory. Sequences
    are kept on the local machine.

    Note: This is not intended to be the endpoint. You should only use temporary files to run the blast search.
    """

    outfmt = [
        "7",
        "qacc",
        "sacc",
        "score",
        "evalue",
        "bitscore",
        "length",
        "nident",
        "gapopen",
        "gaps",
        "qlen",
        "qstart",
        "qend",
        "slen",
        "sstart",
        "send",
        "sstrand",
        "qseq",
        "sseq",
    ]

    blast_config = {"outfmt": '"{0}"'.format(" ".join(outfmt))}

    def __init__(
        self,
        db_name,
        subject_path,
        query_path,
        db_output_directory,
        results_out_path,
        output_formatter=None,
        **additional_config
    ):
        """
        A Blast initializer for running blast searches.

        :param db_name: Name for database file structure. This name will be appended to all db
        files that blast creates.
        :param subject_path: single fsa file containing the subject sequences
        :param query_path: Location of the fasta or genbank file containing the query
        :param db_output_directory: Location to store database related files
        :param results_out_path: Path to store the results.out file. Path can be absolute or relative.
        :param config: Additional configurations to run for the blast search (see
        https://www.ncbi.nlm.nih.gov/books/NBK279682/)
        """

        # add executable to the system path
        blast_wrapper = BlastWrapper()
        if not blast_wrapper.is_installed():
            blast_wrapper.ask_to_install()

        # name of the database
        self.name = db_name

        self.query_path = query_path
        self.subject_path = subject_path
        self._config = {}

        # build the configuration
        if output_formatter is None:
            self.outfmt = BlastBase.outfmt
        else:
            self.outfmt = output_formatter

        # path to the directory holding the blast database
        self.db_output_directory = os.path.abspath(db_output_directory)

        # path to the blast database
        self.db = os.path.join(db_output_directory, db_name)

        # list of expected fields in the result
        self.fields = ()

        # the raw results from running blast from the command line
        self.raw_results = None

        # the parsed results as a dictionary
        self.results = None  #

        # the path to the saved results
        self.results_out_path = os.path.abspath(results_out_path)

    @property
    def config(self):
        out_config = {"outfmt": '"{0}"'.format(" ".join(self.outfmt))}
        out_config.update(self._config)
        return out_config

    def update_config(self, config):
        self._config.update(config)

    def validate_files(self):
        """
        Validate the directories and query files

        :return: None
        :rtype: None
        :raises: PyBlastException if filepaths are invalid or if query sequence has more than one sequence
        """

        def _is_file(myfile):
            return os.path.isfile(os.path.abspath(myfile))

        def _is_dir(mydir):
            return os.path.isdir(os.path.abspath(mydir))

        outdir = os.path.dirname(self.results_out_path)
        errors = []
        for file_ in [self.query_path, self.subject_path]:
            if not _is_file(file_):
                errors.append("File not found: {}".format(file_))
        for dir_ in [outdir, self.db_output_directory]:
            if not _is_dir(dir_):
                errors.append("Directory not found {}".format(dir_))
        if len(errors) > 0:
            raise PyBlastException("\n".join(errors))

        # seq = PySequence.open(self.query_path)
        # if len(seq) == 0:
        #     raise PyBlastException("Query path \"{}\" has no sequences".format(self.query_path))
        # elif len(seq) > 1:
        #     raise PyBlastException("Query path \"{}\" has more than one sequence.".format(self.query_path))

    def create_config(self):
        """Create a configuration dictionary"""
        config_dict = {
            "db": self.db,
            "out": self.results_out_path,
            "query": self.query_path,
        }
        config_dict.update(self.config)
        return config_dict

    def quick_blastn(self):
        """Produce database, run blastn protocol, & parse results """
        self.makedb()
        self.blastn()
        return self.parse_results()

    def quick_blastn_short(self):
        """Produce database, run blastn short protocol, & parse results """
        self.makedb()
        self.blastn_short()
        return self.parse_results()

    def blastn(self):
        return self._run("blastn")

    def blastn_short(self):
        return self._run("blastn", task="blastn-short")

    def _run(self, cmd, **config_opts):
        """Run the blastn using the current configuration"""
        config = self.create_config()
        config.update(config_opts)
        self.run_cmd(cmd, **config)
        with open(self.results_out_path, "rU") as handle:
            self.raw_results = handle.read()
        return self.raw_results

    # Wrapper for the util.run_cmd
    def run_cmd(self, cmd, **kwargs):
        """Wrapper for utils.run_cmd"""
        run_cmd(cmd, **kwargs)

    def _fasta_details(self, path):
        seqs = list(SeqIO.parse(path, format="fasta"))
        tot_bps = sum([len(s) for s in seqs])
        return {"num_sequence": len(seqs), "total_bps": tot_bps}

    def makedb(self, verbose=False):
        """Creates a blastdb from sequences grabbed from the input directory"""
        self.validate_files()

        if verbose:
            query_details = self._fasta_details(self.query_path)
            subject_details = self._fasta_details(self.subject_path)

            print("Making BLAST database from:")
            print(json.dumps(subject_details, indent=2))

            print("Query:")
            print(json.dumps(query_details, indent=2))

        self.run_cmd(
            "makeblastdb",
            dbtype="nucl",
            title=self.name,
            out=self.db,
            **{"in": self.subject_path}
        )
        return self.db

    def _unique_results(self, results):
        return list(
            unique_everseen(
                results,
                key=lambda x: (
                    x["query"]["sequence_id"],
                    x["query"]["start"],
                    x["query"]["end"],
                    x["subject"]["sequence_id"],
                    x["subject"]["start"],
                    x["subject"]["end"],
                ),
            )
        )

    def parse_results(self, delim=","):
        """
        Parses the raw blast result to a JSON

        :param save_as_json: whether to save the JSON. Results will be saved in same directory and name as
        results_out_path but with a .json extension.
        :type save_as_json: bool
        :param delim: delimiter to parse
        :type delim: str
        :return:
        :rtype:
        """
        results = BlastResultParser.raw_results_to_json(self.raw_results, delim=delim)
        results = self._unique_results(results)
        self.results = results
        return self.results

    def get_perfect(self):
        return BlastResultParser.get_perfect(self.results)

    def __str__(self):
        return "{}".format(self.create_config())


class TmpBlast(BlastBase):
    """
    A Blast object that stores the database files in a hidden temporary directory. Use
    "quick_blastn" for returning results as a python object.
    """

    def __init__(self, db_name, subject_path, query_path, **config):
        """
        TmpBlast: A Blast object that stores the database files in a hidden temporary directory.
        :param db_name: Name for database file structure. This name will be appended to all db files that blast creates.
        :param subject_path: Input directory or file containing a list of subjects to align against the query
        :param query_path: Location of the fasta or genbank file containing the query
        :param config: Additional configurations to run for the blast search (see
        https://www.ncbi.nlm.nih.gov/books/NBK279682/)
        """

        db_output_directory = tempfile.mkdtemp()
        fv, out = tempfile.mkstemp(dir=db_output_directory)

        if os.path.isdir(subject_path):
            subject_path = glob_fasta_to_tmpfile(subject_path)

        super(TmpBlast, self).__init__(
            db_name=db_name,
            subject_path=subject_path,
            query_path=query_path,
            db_output_directory=db_output_directory,
            results_out_path=out,
            **config
        )

        self.fields = self.fields + (
            "subject_name",
            "subject_filename",
            "subject_circular",
            "query_name",
            "query_filename",
            "query_circular",
        )

    @classmethod
    def use_test_data(cls):
        """Create a Blast instance using predefined data located in tests"""
        dir_path = os.path.dirname(os.path.realpath(__file__))

        return cls(
            "db",
            os.path.join(dir_path, "..", "..", "tests/data/test_data/db.fsa"),
            os.path.join(dir_path, "..", "..", "tests/data/test_data/query.fsa"),
        )

    def makedb(self):
        super(TmpBlast, self).makedb()


class BioBlast(TmpBlast):
    def __init__(
        self,
        subjects: typing.Sequence[SeqRecord],
        queries: typing.Sequence[SeqRecord],
        span_origin=True,
        **config
    ):

        self.span_origin = span_origin
        self.seq_db = SeqRecordDB()
        subjects = self.add_records(subjects)
        queries = self.add_records(queries)
        self.subjects = subjects
        self.queries = queries
        db_name = str(uuid4())
        subject_path = records_to_tmpfile(subjects)
        query_path = records_to_tmpfile(queries)
        super().__init__(
            db_name=db_name, subject_path=subject_path, query_path=query_path, **config
        )

    def add_records(self, records):
        clean_records(records)

        def copy_record(r):
            return deepcopy(r)

        def pseudocircularize(r):
            r2 = r + r
            r2.name = C.PSEUDOCIRCULAR + "__" + r.name
            r2.id = str(uuid4())
            return r2

        circular = [r for r in records if self.seq_db.is_circular(r)]
        linear = [r for r in records if not self.seq_db.is_circular(r)]
        if self.span_origin:
            keys = self.seq_db.add_many_with_transformations(
                circular, pseudocircularize, C.PSEUDOCIRCULAR
            )
        else:
            keys = self.seq_db.add_many_with_transformations(
                circular, copy_record, C.PSEUDOCIRCULAR
            )
        keys += self.seq_db.add_many_with_transformations(
            linear, copy_record, C.COPY_RECORD
        )
        return self.seq_db.get_many(keys)

    # def _merge_results(self, results):
    #
    #     def mergable(d1, d2):
    #         if d1['end'] == d1['length']:
    #             if d1['circular']:
    #                 return True
    #             else:
    #                 return False
    #         elif d1['end'] + 1 == d2['start']:
    #             return True
    #         return False
    #
    #     def alignment_mergable(v1, v2):
    #         return all([mergable(v1[x], v2[x]) for x in ['query', 'subject']])

    # sorted_by_query_ends = sorted(results, key=lambda x: x['query']['end'])
    # grouped_by_qs = groupby_transform(results,
    #                                   keyfunc=lambda x: x['query']['sequence_id'] + "__" + x['subject']['sequence_id'])
    # for _, group in grouped_by_qs:
    #     grouped_by_query_starts = dict(groupby_transform(group, keyfunc=lambda x: x['query']['start']))
    #     for v1 in group:
    #         e = v1['query']['end']
    #         if e == v1['query']['length']:
    #             qs = grouped_by_query_starts.get(0, list())
    #             for v2 in grouped_by_query_starts.get(0, list()):
    #                 if alignment_mergable(v1, v2):
    #

    def parse_results(self, delim=","):
        parsed_results = BlastResultParser.raw_results_to_json(
            self.raw_results, delim=delim
        )

        # # TODO: resolve with sequence dictionary, resolving pseudocircularized constructs
        for v in parsed_results:
            if v:
                for x in ["query", "subject"]:

                    origin_key = self.seq_db.get_origin_key(v[x]["sequence_id"])
                    record = self.seq_db.get(origin_key)
                    is_circular = self.seq_db.is_circular(record)
                    v[x]["circular"] = is_circular
                    v[x]["name"] = record.name
                    v[x]["origin_key"] = origin_key
                    v[x]["origin_record_id"] = record.id
                    v[x]["origin_sequence_length"] = len(record.seq)
                    # v[x]["length"] = len(record.seq)
                v["meta"]["span_origin"] = self.span_origin

        parsed_results = self._unique_results(parsed_results)
        self.results = parsed_results
        return self.results

    def alignments(self):
        alignments = []
        for align in self.results:
            alignments.append(
                BlastResultParser.alignment_to_seqrecord(align, self.seq_db.records)
            )
        return alignments


class JSONBlast(BioBlast):
    """Object that runs blast starting from JSON inputs and outputs"""

    def __init__(self, subject_json, query_json, span_origin=True, **config):
        """
        Initialize JSONBlast

        :param subject_json: subject sequences as serialized json. Schema may be found in :class:`pyblast.schema.SequenceSchema`
        :type subject_json: list of dict, mapping, or Object
        :param query_json: query sequence as serialized json
        :type query_json: dict, mapping, or Object
        :param preloaded: whether the data is preloaded into the SequenceSchema or not. Practially, this means the "bases"
        for the sequence if refered to as "sequence" in serialized data and as "bases" after data is loaded (preloaded=True)
        :param span_origin: default False. If True, circular sequences will be pseudocircularized for alignment over the origin
        :type span_origin: boolean
        :param config: optional arguments
        :type config: dict
        """
        srecords = JSONParser.JSON_to_SeqRecords(subject_json)
        qrecords = JSONParser.JSON_to_SeqRecords(query_json)
        self.subject_json = subject_json
        self.query_json = query_json
        super().__init__(srecords, qrecords, span_origin=span_origin)

    @classmethod
    def use_test_data(cls):
        """Create a Blast instance using predefined data located in tests"""
        dir_path = os.path.dirname(os.path.realpath(__file__))
        with open(
            os.path.join(dir_path, "..", "..", "tests/data/test_data/templates.json"),
            "r",
        ) as f:
            subject = json.load(f)
        with open(
            os.path.join(dir_path, "..", "..", "tests/data/test_data/query.json"), "r"
        ) as f:
            query = json.load(f)
        return cls(subject_json=subject, query_json=query)



"""utils"""

from .cmd import run_cmd
from Bio.Alphabet import generic_dna
from Bio.Seq import Seq
from glob import glob
from os.path import join
from Bio import SeqIO
import tempfile
from Bio.SeqFeature import FeatureLocation, CompoundLocation
from pyblast.constants import Constants as C


def reverse_complement(seq_str):
    return str(Seq(seq_str, generic_dna).reverse_complement())


def complement(seq_str):
    return str(Seq(seq_str, generic_dna).complement())


def new_feature_location(start, end, length, strand):
    """Makes a FeatureLocation. If necessary, makes CompoundLocation."""
    if start > end:
        if length is None:
            raise ValueError(
                "A length must be provided to create a feature with start > end."
            )
        f1 = FeatureLocation(start, length, strand)
        f2 = FeatureLocation(1, end, strand)
        if strand == -1:
            location = CompoundLocation([f2, f1])
        else:
            location = CompoundLocation([f1, f2])
    else:
        location = FeatureLocation(start, end, strand=strand)
    return location


def records_to_tmpfile(records):
    fd, tmp_path_handle = tempfile.mkstemp(suffix=".fasta")
    SeqIO.write(records, tmp_path_handle, format="fasta")
    return tmp_path_handle


def glob_fasta_to_tmpfile(dir):
    fasta_files = glob(join(dir, "*.fsa"))
    fasta_files += glob(join(dir, "*.fasta"))
    records = []
    for fsa in fasta_files:
        records += list(SeqIO.parse(fsa, "fasta"))
    return records_to_tmpfile(records)


def load_glob(path, format, recursive=False):
    records = []
    for f in glob(path, recursive=recursive):
        records += list(SeqIO.parse(f, format=format))
    return records


def load_fasta_glob(path, recursive=False):
    return load_glob(path, "fasta", recursive=recursive)


def load_genbank_glob(path, recursive=False):
    records = load_glob(path, "genbank", recursive=recursive)
    for r in records:
        if "topology" in r.annotations and r.annotations["topology"] == "circular":
            r.annotations["circular"] = True
        else:
            r.annotations["circular"] = False
    return records


def clean_records(records):
    for r in records:
        r.features = [f for f in r.features if f.location is not None]


def is_circular(r):
    annotations = {k.lower(): v for k, v in r.annotations.items()}
    return (
        annotations.get(C.CIRCULAR.lower(), False) is True
        or annotations.get(C.LINEAR.lower(), True) is False
        or annotations.get(C.TOPOLOGY.lower(), C.LINEAR).lower() == C.CIRCULAR.lower()
    )


def make_linear(records):
    for r in records:
        if C.CIRCULAR in r.annotations:
            r.annotations[C.CIRCULAR] = False
        elif C.LINEAR in r.annotations:
            r.annotations[C.LINEAR] = True
        else:
            r.annotations[C.TOPOLOGY] = C.LINEAR

    return records


def make_circular(records):
    for r in records:
        if C.CIRCULAR in r.annotations:
            r.annotations[C.CIRCULAR] = True
        elif C.LINEAR in r.annotations:
            r.annotations[C.LINEAR] = False
        else:
            r.annotations[C.TOPOLOGY] = C.CIRCULAR
    return records
